unit colorize1;

interface

function colorize_1(maxnode: integer; even,numc: smallint): integer; stdcall;
function colorize_2(maxnode: integer; even: smallint): integer; stdcall;

type
  Tnode=record
          color,lastlink: integer;
          weight: double;
          // are only used in Tcolorize2
          unmarkedneighbors,notaround,pref: integer;
          possiblecolor: boolean;
        end;
  Tcolorize=class(Tobject)
            private
              nodes: array of Tnode;
              nodelink: array of array of integer;
              colorweight: array of double;
              error: double;
            protected
              procedure init;
              procedure countweight;
              procedure exit1;
              procedure exit2;
            public
              maxnode,maxcolor: integer;
              even: boolean;
            end;
  Tcolorize1=class(Tcolorize)
             private
               index: array of integer;
               function AssignOneColor(const rgn_i: Integer): Boolean;
             public
               function Execute: integer;
             end;
  Tcolorize2=class(Tcolorize)
             public
               function Execute: integer;
             end;

implementation

uses
  sysutils;

function colorize_1;
var
  col: Tcolorize1;
begin
  col:= Tcolorize1.create;
  col.maxnode:= maxnode;
  if numc in [4,5,6] then
    col.maxcolor:= numc
  else
    col.maxcolor:= 5;
  if even=1 then col.even:= true else col.even:= false;
  result:= col.execute;
  col.free;
end;

function colorize_2;
var
  col: Tcolorize2;
begin
  col:= Tcolorize2.create;
  col.maxnode:= maxnode;
  if even=1 then col.even:= true else col.even:= false;
  result:= col.execute;
  col.free;
end;

procedure Tcolorize.countweight;
var
  i: integer;
begin
  for i:= 1 to maxcolor do colorweight[i]:= 0;
  for i:= 1 to maxnode do with nodes[i] do colorweight[color]:= colorweight[color]+weight;
  error:= 0;
  for i:= 1 to maxcolor do error:= error+abs(1/maxcolor-colorweight[i]);
end;

procedure Tcolorize.init;
var
  f: textfile;
  a,b,i: integer;
  totweight: double;
begin
  // initialization
  setlength(nodelink,maxnode+1,5);
  setlength(nodes,maxnode+1);
  for i:= 1 to maxnode do with nodes[i] do
  begin
    lastlink:= 0;
    color:= 0;
  end;

  // read input
  assignfile(f,'~graph.txt');
  reset(f);
  while not eof(f) do
  begin
    readln(f,a,b);

    inc(nodes[a].lastlink);
    if high(nodelink[a])<nodes[a].lastlink
      then setlength(nodelink[a],nodes[a].lastlink+5);
    nodelink[a,nodes[a].lastlink]:= b;

    inc(nodes[b].lastlink);
    if high(nodelink[b])<nodes[b].lastlink
      then setlength(nodelink[b],nodes[b].lastlink+5);
    nodelink[b,nodes[b].lastlink]:= a;
  end;
  closefile(f);
  deletefile('~graph.txt');

  // read weights
  assignfile(f,'~weight.txt');
  reset(f);
  for i:= 1 to maxnode do readln(f,nodes[i].weight);
  closefile(f);
  deletefile('~weight.txt');

  // find total weight and normalize
  totweight:= 0;
  for i:= 1 to maxnode do totweight:= totweight+nodes[i].weight;
  for i:= 1 to maxnode do nodes[i].weight:= nodes[i].weight/totweight;
end;

procedure Tcolorize.exit1;
var
  f: textfile;
  col_bef,i,j,mink,k: integer;
  changed,possible: boolean;
  err_bef,minweight: double;
begin
  if even then
  begin
    setlength(colorweight,maxcolor+1);
    // modification according to weights
    countweight;
    repeat
      changed:= false;
      for i:= 1 to maxnode do
      begin
        // too much of this color at node?
        if colorweight[nodes[i].color]>1/maxcolor then
        begin
          // try to see if one of the less represented colors are possible
          minweight:= 1;
          mink:= 0;
          for k:= 1 to maxcolor do if colorweight[k]<1/maxcolor then
          begin
            possible:= true;
            for j:= 1 to nodes[i].lastlink do
              if nodes[nodelink[i,j]].color=k then begin possible:= false; break; end;
            if possible and (colorweight[k]<minweight) then
            begin
              minweight:= colorweight[k];
              mink:= k;
            end;
          end;
          if mink<>0 then
          begin
            err_bef:= error;
            col_bef:= nodes[i].color;
            nodes[i].color:= mink;
            countweight;
            if err_bef<error then      // was it better before ?
            begin
              nodes[i].color:= col_bef;
              countweight;
            end else changed:= true;   // if not, we keep the changes
          end;
        end;
      end;
    until not changed;
    finalize(colorweight);
  end;

  assignfile(f,'~out.txt');
  rewrite(f);
  for i:= 1 to maxnode do writeln(f,nodes[i].color);
  closefile(f);
end;

procedure Tcolorize.exit2;
begin
  finalize(nodelink);
  finalize(nodes);
end;

{ Tcolorize1 }

function Tcolorize1.AssignOneColor(const rgn_i: Integer): Boolean;
var
  id,next,color,nbr_i: Integer;
  stop: boolean;
begin
  id:= index[rgn_i];
  for color:= 1 to maxcolor do
  begin
    stop:= false;
    for nbr_i:= 1 to nodes[id].lastlink do
    if nodes[nodelink[id,nbr_i]].color=color then
    begin
      stop:= true;
      break;
    end;
    if not stop then
    begin
      nodes[id].color:= color;
      next:= rgn_i+1;
      if (next=maxnode+1) or AssignOneColor(next) then
      begin
        result:= true;
        exit;
      end;
    end;
  end;
  nodes[id].color:= 0;
  result:= false;
end;

function Tcolorize1.Execute;
var
  i,j,mindex,tmp: integer;
begin
  init;

  // initialization
  setlength(index,maxnode+1);
  for i:= 1 to maxnode do index[i]:= i;

  i:= 1;
  while i<=maxnode do
  begin
    mindex:= i;
    j:= i+1;
    while j<=maxnode-1 do
    begin
      if nodes[index[j]].lastlink>nodes[index[mindex]].lastlink then mindex:= j;
      inc(j);
    end;
    tmp:= index[mindex]; index[mindex]:= index[i]; index[i]:= tmp;
    i:= i+1;
  end;

  if AssignOneColor(1) then
  begin
    exit1;
    result:= maxcolor;
  end else result:= 0;

  finalize(index);
  exit2;
end;

{ Tcolorize2 }

function Tcolorize2.Execute;
var
  uncolored,x,max_val,i,j,k,nb,gb: integer;
  first: boolean;
begin
  init;

  // initialization
  for i:= 1 to maxnode do with nodes[i] do
  begin
    notaround:= 0;
    pref:= 0;
  end;

  // the number of unmarked neighbors is the same as the number of neighbors
  for i:= 1 to maxnode do with nodes[i] do unmarkedneighbors:= lastlink;

  // all nodes are uncolored
  uncolored:= maxnode;

  maxcolor:= 0;
  repeat
    inc(maxcolor);
    for i:= 1 to maxnode do nodes[i].possiblecolor:= (nodes[i].color=0);
    repeat
      // test if there are any nodes left to color with the current color
      i:= 1;
      while not nodes[i].possiblecolor and (i<maxnode) do inc(i);
      if not nodes[i].possiblecolor then break;

      // find biggest pref
      first:= true;
      for i:= 1 to maxnode do with nodes[i] do if possiblecolor then
      begin
        if first or (max_val<pref*10000+notaround*100+unmarkedneighbors) then
        begin
          max_val:= pref*10000+notaround*100+unmarkedneighbors;
          x:= i;
          first:= false;
        end;
      end;

      nodes[x].color:= maxcolor;
      nodes[x].possiblecolor:= false;
      dec(uncolored);
      for i:= 1 to nodes[x].lastlink do
      begin
        nb:= nodelink[x,i];
        if nodes[nb].color=0 then
        begin
          dec(nodes[nb].unmarkedneighbors);
          nodes[nb].possiblecolor:= false;
        end;
        for j:= 1 to nodes[nb].lastlink do
        begin
          gb:= nodelink[nb,j];
          inc(nodes[gb].pref);
          nodes[gb].notaround:= 0;
          for k:= 1 to nodes[gb].lastlink do
            if not nodes[nodelink[gb,k]].possiblecolor then
              inc(nodes[gb].notaround);
        end;
      end;
    until false;
    for i:= 1 to maxnode do with nodes[i] do if color=0 then
    begin
      notaround:= 0;
      pref:= 0;
    end;
  until uncolored=0;

  exit1;
  exit2;
  result:= maxcolor;
end;

end.
