unit nearestn;

interface

  // nearest:     number of nearest objects (max 100)
  // coordtype:   0 = cartesian, 1 = latlong (km), 2 = latlong (miles)
  // outputmode:  1 means many fields, 2 means many records
  // maxdist:     0 or a maximum value
  
function NearestNodes(nearest,coordtypeV,outputmode: integer; maxdist: double): integer; stdcall;

implementation

uses
  math,sysutils,sec;

function NearestNodes;
var
  inp: textfile;
  outp: textfile;
  X,Y,dist,outdist: array of double;
  x1,y1,act_dist,maxdist_found: double;
  i,j,k,l,count,maxj,last,method: integer;
  nodex,outindex: array of integer;

  procedure QuickSort(iLo, iHi: longInt);
  var
    Lo,Hi,T: longInt;
    Mid: double;
  begin
    Lo:= iLo;
    Hi:= iHi;
    Mid:= dist[nodex[(Lo + Hi) div 2]];
    repeat
      while dist[nodex[Lo]]<Mid do Inc(Lo);
      while dist[nodex[Hi]]>Mid do Dec(Hi);
      if Lo<=Hi then
      begin
        T:= nodex[Lo]; nodex[Lo]:= nodex[Hi]; nodex[Hi]:= T;
        Inc(Lo);
        Dec(Hi);
      end;
    until Lo>Hi;
    if Hi>iLo then QuickSort(iLo,Hi);
    if Lo<iHi then QuickSort(Lo,iHi);
  end;

begin
  coordtype:= coordtypeV;
  result:= 0;
  method:= 0;
  try try
    assignfile(inp,'tmp1.txt');
    reset(inp);
    count:= 0;
    while not eof(inp) do begin readln(inp); inc(count); end;
    closefile(inp);

    if nearest>count then nearest:= count;
    if nearest>100 then nearest:= 100;
    if (nearest=0) and (maxdist<>0) then
    begin
      if outputmode=2 then nearest:= count else nearest:= 100;
    end;
    if nearest=0 then nearest:= 1;

    // decide calculation method
    // method 1: Simple select
    // method 2: Bubble sort like method (the normal)
    // method 3: Quicksort
    if nearest=1 then method:= 1 else
    if ((coordtypeV=0) and (sqrt(count)/nearest>0.26)) or
       ((coordtypeV<>0) and (sqrt(count)/nearest>0.23)) then method:= 2 else
      method:= 3;

    setlength(x,count+1);
    setlength(y,count+1);
    if method=2 then setlength(outdist,nearest+1);
    if method=2 then setlength(outindex,nearest+1);
    if method=3 then setlength(dist,count+1);
    if method=3 then setlength(nodex,count+1);

    // read x/y coordinates for "stores"
    reset(inp);
    for i:= 1 to count do readln(inp,x[i],y[i]);
    closefile(inp);

    // change latlong into radians
    if coordtypeV in [1,2] then
    begin
      for i:= 1 to count do x[i]:= x[i]/180*pi;
      for i:= 1 to count do y[i]:= y[i]/180*pi;
    end;

    // prepare header for output file
    assignfile(outp,'nearest.txt');
    rewrite(outp);
    if outputmode=1 then
    begin
      write(outp,'OrgID');
      for j:= 1 to nearest do write(outp,' rowID',inttostr(j),' dist',inttostr(j));
      writeln(outp);
    end else
    begin
      writeln(outp,'OrgID Nearest rowID0 Dist0');
    end;

    //read x/y coordinates for "customers"
    assignfile(inp,'tmp2.txt');
    reset(inp);
    i:= 1;
    while not eof(inp) do
    begin
      readln(inp,x1,y1);
      if coordtypeV in [1,2] then
      begin
        x1:= x1/180*pi;
        y1:= y1/180*pi;
      end;

      // select just the nearest
      case method of
      1: begin
           maxdist_found:= 1e38;
           maxj:= 0;
           for j:= 1 to count do
           if (AirDist2(y1,y[j])<maxdist_found) and (AirDist1(x1,x[j],max(abs(y1),abs(y[j])))<maxdist_found) then
           begin
             act_dist:= AirDistPos(x1,x[j],y1,y[j]);
             if act_dist<maxdist_found then
             begin
               maxdist_found:= act_dist;
               maxj:= j;
             end;
           end;
           // write result
           if outputmode=1 then
           begin
             if (maxdist>0) and (maxdist_found>maxdist) then
               writeln(outp,i,' 0 0') else
               writeln(outp,i,' ',maxj,' ',maxdist_found);
           end else
           begin
             if (maxdist>0) and (maxdist_found>maxdist) then
               writeln(outp,i,' 1 0 0') else
               writeln(outp,i,' 1 ',maxj,' ',maxdist_found);
           end;
         end;
      2: begin
           for j:= 1 to nearest do outdist[j]:= 1e38;
           outindex[1]:= 0;
           last:= 1;
           for j:= 1 to count do
           if (AirDist2(y1,y[j])<outdist[last]) and (AirDist1(x1,x[j],max(abs(y1),abs(y[j])))<outdist[last]) then
           begin
             act_dist:= AirDistPos(x1,x[j],y1,y[j]);
             if act_dist<outdist[last] then
             begin
               k:= last;
               while (k<>0) and (outdist[k]>act_dist) do dec(k);
               for l:= min(last,nearest-1) downto k+1 do
               begin
                 outdist[l+1]:= outdist[l];
                 outindex[l+1]:= outindex[l];
               end;
               outdist[k+1]:= act_dist;
               outindex[k+1]:= j;
               if last<>nearest then inc(last);
             end;
           end;
           // write result
           if outputmode=1 then
           begin
             write(outp,i);
             for j:= 1 to nearest do
             begin
               if maxdist=0 then write(outp,' ',outindex[j],' ',outdist[j]) else
               begin
                 if outdist[j]<=maxdist then
                   write(outp,' ',outindex[j],' ',outdist[j])
                 else write(outp,' 0 0');
               end;
             end;
             writeln(outp);
           end else
           begin
             for j:= 1 to nearest do
             begin
               if maxdist=0 then writeln(outp,i,' ',j,' ',outindex[j],' ',outdist[j]) else
               begin
                 if outdist[j]<=maxdist then
                   writeln(outp,i,' ',j,' ',outindex[j],' ',outdist[j]) else break;
               end;
             end;
           end;
         end;
      3: begin
           for j:= 1 to count do dist[j]:= AirDistPos(x1,x[j],y1,y[j]);
           for j:= 1 to count do nodex[j]:= j;
           quicksort(1,count);
           if outputmode=1 then
           begin
             write(outp,i);
             for j:= 1 to nearest do
             begin
               if maxdist=0 then write(outp,' ',nodex[j],' ',dist[nodex[j]]) else
               begin
                 if dist[nodex[j]]<=maxdist then
                   write(outp,' ',nodex[j],' ',dist[nodex[j]])
                 else write(outp,' 0 0');
               end;
             end;
             writeln(outp);
           end else
           begin
             for j:= 1 to nearest do
             begin
               if maxdist=0 then writeln(outp,i,' ',j,' ',nodex[j],' ',dist[nodex[j]]) else
               begin
                 if dist[nodex[j]]<=maxdist then
                   writeln(outp,i,' ',j,' ',nodex[j],' ',dist[nodex[j]]) else break;
               end;
             end;
           end;
         end;
      end;
      inc(i);
    end;
    closefile(inp);
    closefile(outp);
  except
    result:= 1;
  end;
  finally
    finalize(x);
    finalize(y);
    if method=2 then finalize(outdist);
    if method=2 then finalize(outindex);
    if method=3 then finalize(nodex);
    if method=3 then finalize(dist);
  end;
end;

end.
